# system imports
from .imports import *


### convenient objs
def kwargs_key(kwargs,bad_keys={'num_proc','progress','desc'}):
    return hashstr(str(sorted(
        (k,v)
        for k,v in kwargs.items()
        if k not in bad_keys
    )))




def Verse(txt,**kwargs):
    kwargs={**dict(verse=True), **kwargs}
    return Text(txt,**kwargs)
def Poem(txt,**kwargs):
    kwargs={**dict(verse=True), **kwargs}
    return Text(txt,**kwargs)
def Prose(txt,**kwargs):
    kwargs={**dict(prose=True), **kwargs}
    return Text(txt,**kwargs)
def FreeVerse(txt,**kwargs):
    kwargs={**dict(linebreaks=True,phrasebreaks=True), **kwargs}
    return Text(txt,**kwargs)
    
### texts
class Text(object):
    def __init__(self,
            txt_or_fn_or_scandf,
            **kwargs):
        self.fn=None
        self.txt=None
        self.scandf=None
        self._scans={}
        self._parses={}
        self._num_lines=None
        self._num_stanzas=None
        self.infod={}
        
        self.kwargs=kwargs
        for k,v in self.kwargs.items(): setattr(self,k,v)
        self.kwargs_key=kwargs_key(self.kwargs)
        
        # load data
        if type(txt_or_fn_or_scandf)==str:
            if os.path.exists(txt_or_fn_or_scandf):
                self.fn=txt_or_fn_or_scandf
                self.txt=to_txt(self.fn)
            else:
                self.fn=None
                self.txt=txt_or_fn_or_scandf
        elif type(txt_or_fn_or_scandf)==pd.DataFrame:
            self.fn=None
            self.scandf=self._scans[self.kwargs_key]=scandf=txt_or_fn_or_scandf
            self.txt=to_txt_from_scandf(scandf)
        else:
            raise Exception(f'Input not recognised: {type(txt_or_fn_or_scandf)}')
    
    
    
    def __repr__(self):
        o=self.txt.split('\n\n')[0] if self.txt is not None else ""
        o='\t' + '\n\t'.join(l for l in o.split('\n'))
        o=f'''<cadence.Text: {self.first_line} ({self.num_stanzas} stanza{"s" if self.num_stanzas>1 else ""}, {self.num_lines} line{"s" if self.num_lines>1 else ""})>'''.strip()
        #o='\n'.join(l.strip() for l in o.split('\n'))
        return o
    
    @property
    def first_line(self):
        from string import punctuation
        return self.txt.strip().split('\n')[0].strip(punctuation).strip()
    @property
    def first_stanza(self):
        return self.txt.strip().split('\n\n')[0].strip()
    
    
    @property
    def num_lines(self):
        if self._num_lines is None:
            self._num_lines=get_num_lines(
                self.txt if self.scandf is None else self.scandf
            )
        return self._num_lines
    @property
    def num_stanzas(self):
        if self._num_stanzas is None:
            self._num_stanzas=get_num_stanzas(
                self.txt if self.scandf is None else self.scandf
            )
        return self._num_stanzas
    
    
    def scan(self, force=False, **kwargs):
        kwargs={**self.kwargs, **kwargs}
        key=kwargs_key(kwargs)
        if force or not key in self._scans:
            self._scans[key]=scan(self.txt,**kwargs)
        return self._scans[key]
    
    def parse(self,
            force=False,
            verbose=True,
            only_best=False,
            only_unbounded=True,
            **kwargs):
        kwargs['verbose']=verbose
        kwargs_line={**self.kwargs, **kwargs, **{'by_syll':False}}
        kwargs_syll={**self.kwargs, **kwargs, **{'by_syll':True}}
        key_line=kwargs_key(kwargs_line)
        key_syll=kwargs_key(kwargs_syll)
        if force or not (key_syll in self._parses) or (not key_line in self._parses):
            self._parses[key_syll]=parse(self.txt, **kwargs_syll)
            self._parses[key_line]=to_lines(self._parses[key_syll])
        elif kwargs.get('verbose',True):
            for li,linedf in sorted(self._parses[key_syll].reset_index().groupby(['stanza_i','line_i'])):
                display(show_parse(linedf))

        self.infod=info_parses(self._parses[key_line])
        if verbose: printm(show_info_parses(self.infod))
                
        
    def parses(self,
            force=True,
            only_best=False,
            only_unbounded=True,
            **kwargs):
        kwargs={**self.kwargs, **kwargs}
        kwargs_line={**self.kwargs, **kwargs, **{'by_syll':False}}
        kwargs_syll={**self.kwargs, **kwargs, **{'by_syll':True}}
        key_line=kwargs_key(kwargs_line)
        key_syll=kwargs_key(kwargs_syll)
        key=key_syll if kwargs.get('by_syll') else key_line
        if not key in self._parses:
            self.parse(force=force, **kwargs)
        if not key in self._parses: return
        
        odf=self._parses[key]
        if only_unbounded and ('parse_is_bounded' in set(odf.index.names) or 'parse_is_bounded' in set(odf.columns)):
            #odf=odf[odf.parse_is_bounded==False]
            odf=odf.query('parse_is_bounded==False')
        if only_best and ('parse_rank' in set(odf.index.names) or 'parse_rank' in set(odf.columns)):
            odf=odf.query('parse_rank==1')
        
        return odf

    def best_parses(self, force=False, **kwargs):
        return self.parses(force=force,only_best=True,**kwargs)
    def all_parses(self, force=False,**kwargs):
        return self.parses(force=force,only_best=False,only_unbounded=False,**kwargs)
    def unbounded_parses(self, force=False,**kwargs):
        return self.parses(force=force,only_best=False,only_unbounded=True,**kwargs)


from ..imports import *

def split_punct(tok):
    toks=tokenize_agnostic(tok)
    wordil=[]
    
    try:
        for i,x in enumerate(toks):
            if any(y.isalpha() for y in x): wordil+=[i]
        pref=''.join([
            x
            for i,x in enumerate(toks)
            if i<wordil[-1]
        ])
        suf=''.join([
            x
            for i,x in enumerate(toks)
            if i>wordil[-1]
        ])
        words=''.join(
            toks[wordil[0] : wordil[-1]+1]
        ) if wordil else ''
        return (pref,words,suf)
    except IndexError:
        return ('',tok,'')

def zero_punc(token):
    return ''.join(x for x in token if x.isalpha())

def tokenize_agnostic(txt):
    return re.findall(r"[\w']+|[.,!?; -—–\n]", txt)
    
def tokenize_fast(line,lower=False):
	line = line.lower() if lower else line
	import re
	# tokenize using reg ex (fast)
	tokens = re.findall("[A-Z]{2,}(?![a-z])|[A-Z][a-z]+(?=[A-Z])|[\'\w\-]+",line)
	# remove punctuation on either end
	from string import punctuation
	tokens = [tok.strip(punctuation) for tok in tokens]
	# make sure each thing in list isn't empty
	tokens = [tok for tok in tokens if tok]
	return tokens

# tokenize
def tokenize_nltk(txt,lower=False):
	# lowercase
	txt_l = txt.lower() if lower else txt
	# use nltk
	tokens = nltk.word_tokenize(txt_l)
	# weed out punctuation
	tokens = [
		tok
		for tok in tokens
		if tok[0].isalpha()
	]
	# return
	return tokens

def tokenize_nice(xstr,starters=set("([‘“"),**kwargs):
    l=tokenize_agnostic(xstr)
    l2=[]
    addback=False
    for x in l:
        if not x:#.strip():
            pass
        elif x[0].isalpha():
            if addback and len(l2):
                l2[-1]+=x
                addback=False
            else:
                l2+=[x]
        else:
            if x in starters:
                l2+=[x]
                addback=True
            elif len(l2):
                l2[-1]+=x
            else:
                l2+=[x]
                addback=True
    return l2


# def tokenize(txt,*x,**y):
# 	return tokenize_fast(txt,*x,**y)




def subfinder(mylist, pattern):
    matches = []
    for i in range(len(mylist)):
        if mylist[i] == pattern[0] and mylist[i:i+len(pattern)] == pattern:
            matches.append(pattern)
    return matches

DTOK_TREEBANK=None
def detokenize_treebank(x):
    global DTOK_TREEBANK
    if DTOK_TREEBANK is None:
        from nltk.tokenize.treebank import TreebankWordDetokenizer
        DTOK_TREEBANK = TreebankWordDetokenizer()
    return DTOK_TREEBANK.detokenize(x)

def detokenize(x):
    x=detokenize_treebank(x)
    return x

def setindex(df,key=LINEKEY,badcols={'index','level_0'},sort=True):
    df = df[set(df.columns) - set(df.index.names)]
    df = resetindex(df)

    cols=[]
    for x in key:
        if not x in set(cols) and x in set(df.columns):
            cols.append(x)

    odf=df.set_index(cols)
    if sort: odf=odf.sort_index()
    odf=odf[[col for col in odf.columns if not col in badcols and col not in key]]

    constraints = [c for c in odf.columns if c.startswith('*')]
    resortcols = ['*total'] + [c for c in constraints if c!='*total']
    resortcols+= list(sorted([c for c in odf.columns if c not in set(constraints)]))
    return odf[[c for c in resortcols if c in odf.columns]]

def resetindex(df,badcols={'level_0','index'},**y):
    cols=set(df.columns)
    inds=set(df.index.names)
    both=cols&inds
    # print(cols)
    # print(inds)
    # print(both)
    newdf=df[cols - both - badcols].reset_index(**y)
    # print(newdf.columns)
    # print()
    return newdf


def occurrences(string, sub):
    count = start = 0
    while True:
        start = string.find(sub, start) + 1
        if start > 0:
            count+=1
        else:
            return count

import numpy as np
def rolling_slices(df,window_len=3,incl_empty=True,keep_last_keys=[]):
    nrad=(window_len - 1)//2
    for i in range(len(df)):
        mini = i-nrad
        maxi = i+nrad
        empty_row=pd.Series(dict((k,np.nan) for k in df.columns))
        for xx in keep_last_keys: empty_row[xx]=df.iloc[0][xx]
        irows=[]
        inames=[]
        for ii in range(mini,maxi+1):
            inames+=[ii]
            if ii<0 or ii>=len(df):
                if incl_empty:

                    irows+=[empty_row]
            else:
                irows+=[df.iloc[ii]]
        yield pd.DataFrame(irows,index=inames)
def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

# def product(*args):
# 	if not args:
# 		return iter(((),)) # yield tuple()
# 	return (items + (item,)
# 		for items in product(*args[:-1]) for item in args[-1])


def hashstr(*x):
    import hashlib
    return hashlib.sha224(str(x).encode('utf-8')).hexdigest()

def pmap_do(inp):
    func,obj,args,kwargs = inp
    return func(obj,*args,**kwargs)

def pmap_iter(func, objs, args=[], kwargs={}, num_proc=DEFAULT_NUM_PROC, use_threads=False, progress=True, desc=None, position=0, leave=True, **y):
    """
    Yields results of func(obj) for each obj in objs
    Uses multiprocessing.Pool(num_proc) for parallelism.
    If use_threads, use ThreadPool instead of Pool.
    Results in any order.
    """

    # imports
    import multiprocessing as mp
    from tqdm import tqdm

    # check num proc
    num_cpu = mp.cpu_count()
    if num_proc>num_cpu: num_proc=num_cpu

    # if parallel
    if not desc: desc=f'Mapping {func.__name__}()'
    if desc: desc=f'{desc} [x{num_proc}]'
    if num_proc>1 and len(objs)>1:

        # real objects
        objects = [(func,obj,args,kwargs) for obj in objs]

        # create pool
        pool=mp.Pool(num_proc) if not use_threads else mp.pool.ThreadPool(num_proc)

        # yield iter
        iterr = pool.imap(pmap_do, objects)
        iterr=tqdm(
            iterr,
            total=len(objects),
            desc=desc,
            disable=not progress,
            leave=leave,
            position=position
        )
        yield from iterr

        # Close the pool?
        pool.close()
        pool.join()
    else:
        # yield
        iterr=tqdm(
            objs,
            desc=desc,
            disable=not progress,
            leave=leave,
            position=position
        )
        for obj in iterr:
            yield func(obj,*args,**kwargs)

def pmap(*x,**y):
    """
    Non iterator version of pmap_iter
    """
    # return as list
    return list(pmap_iter(*x,**y))

def index_by_truth(x):
    i=0
    for y in x:
        if y:
            yield i
            i+=1
        else:
            yield np.nan


def do_pmap_group(obj,*args,**kwargs):
    import pandas as pd
    import types
    func,group_df,group_key,group_name = obj

    if type(group_name) not in {list,tuple}:group_name=[group_name]
    if type(group_df)==str: group_df=pd.read_pickle(group_df)
    out=func(group_df,*args,**kwargs)
    if isinstance(out, types.GeneratorType):
        out=[x for x in out]
    if type(out)==list:
        try:
            out=[x for x in out if type(x) in {pd.DataFrame, pd.Series}]
            out=pd.concat(out)
        except ValueError:
            return out
    if type(out)==pd.DataFrame:
        for x,y in zip(group_key,group_name): out[x]=y
    return out

def joindfs(a,b):
    bcols = set(b.columns) - set(a.columns)
    return a.join(b[bcols])


def hashstr(x):
	import hashlib
	return hashlib.sha224(str(x).encode('utf-8')).hexdigest()


# Use sqlite dictionary
def get_db(
        prefix,
        mode='c',
        folders=[],
        autocommit=False,
        ):
    if mode=='w': autocommit=True
    ofnfn=os.path.join(PATH_DATA, f'db.{prefix}.sqlite')
    odir=os.path.dirname(ofnfn)
    if not os.path.exists(odir): os.makedirs(odir)
    if not os.path.exists(ofnfn): mode='c'
    return SqliteDict(
        ofnfn,
        tablename='data',
        autocommit=autocommit,
        flag='r' if mode=='r' else 'c',
        timeout=30
    )



def slices(l,n,strict=True):
    o=[]
    for x in l:
        o.append(x)
        if len(o)>n: o.pop(0)
        if not strict or len(o)==n:
            yield list(o)
def apply_combos(df,group1,group2,combo_key='combo_i'):
    # combo of indices?
    combo_opts = [
        [x for ii,x in grp.groupby(group2)]
        for i,grp in df.groupby(group1)
    ]

    # poss
    for combo in product(*combo_opts):
        if not len(combo): continue
        yield pd.concat(combo)# if len(combo) else pd.DataFrame()
def pmap_iter_groups(func,df_grouped,use_cache=False,num_proc=DEFAULT_NUM_PROC,iter=False,**attrs):
    import os,tempfile,pandas as pd
    from tqdm import tqdm


    # get index/groupby col name(s)
    group_key=df_grouped.grouper.names
    # if not using cache
    # if not use_cache or attrs.get('num_proc',1)<2:
    if not use_cache or len(df_grouped)<2 or num_proc<2:
        objs=[
            (func,group_df,group_key,group_name)
            for group_name,group_df in df_grouped
        ]
    else:
        objs=[]
        tmpdir=tempfile.mkdtemp()
        # for i,(group_name,group_df) in enumerate(tqdm(list(df_grouped),desc='Preparing input')):
        for i,(group_name,group_df) in enumerate(df_grouped):
            tmp_path = os.path.join(tmpdir, str(i)+'.pkl')
            # print([i,group_name,tmp_path,group_df])
            group_df.to_pickle(tmp_path)
            objs+=[(func,tmp_path,group_key,group_name)]

    # desc?
    if not attrs.get('desc'): attrs['desc']=f'Mapping {func.__name__}'

    #iterfunc = pmap if not iter else pmap_iter
    #return pd.concat(iterfunc) if not iter else
    return pmap_iter(
        do_pmap_group,
        objs,
        num_proc=num_proc,
        **attrs
    )
def pmap_groups(*x,**y):
    res = list(pmap_iter_groups(*x,**y))
    # print([type(x) for x in res])
    # for y in res[-1]:
        # print(type(y), y)
    resl = []
    for x in res:
        if type(x)==list:
            for y in x:
                resl.append(y)
        else:
            resl.append(x)
    return pd.concat(resl) if len(resl) else pd.DataFrame()


def check_basic_config():
    # check basic config
    try:
        if not os.path.exists(PATH_HOME): os.makedirs(PATH_HOME)
    except Exception:
        pass
    if not os.path.exists(PATH_DATA): os.makedirs(PATH_DATA)
    if not os.path.exists(os.path.join(PATH_DATA,'en')):
        zipfn=os.path.join(PATH_HOME,'data_cadence.zip')
        download(DATA_URL, zipfn)
        unzip(zipfn, PATH_HOME)
    try:
        nltk.word_tokenize('testing')
    except LookupError:
        nltk.download('punkt')
    try:
        nltk.corpus.stopwords.words('english')
    except LookupError:
        nltk.download('stopwords')




### utils
def printm(x):
    try:
        from IPython.display import Markdown
        return display(Markdown(x))
    except ImportError:
        print(x)



def download_wget(url, save_to, **attrs):
    import wget
    save_to_dir,save_to_fn=os.path.split(save_to)
    if save_to_dir:
        if not os.path.exists(save_to_dir): os.makedirs(save_to_dir)
        os.chdir(save_to_dir)
    fn=wget.download(url,bar=wget.bar_adaptive)
    os.rename(fn,save_to_fn)
    # print('\n>> saved:',save_to)

def download(url,save_to,force=False,desc=''):
    here=os.getcwd()
    download_wget(url,save_to,desc=desc)
    os.chdir(here)




def unzip(zipfn, dest='.', flatten=False, overwrite=False, replace_in_filenames={},desc='',progress=True):
    from zipfile import ZipFile
    from tqdm import tqdm

    # Open your .zip file
    if not desc: desc=f'Extracting {os.path.basename(zipfn)}'
    with ZipFile(zipfn) as zip_file:
        namelist=zip_file.namelist()

        # Loop over each file
        iterr=tqdm(iterable=namelist, total=len(namelist),desc=desc) if progress else namelist
        for member in iterr:
            # Extract each file to another directory
            # If you want to extract to current working directory, don't specify path
            filename = os.path.basename(member)
            if not filename: continue
            target_fnfn = os.path.join(dest,member) if not flatten else os.path.join(dest,filename)
            for k,v in replace_in_filenames.items(): target_fnfn = target_fnfn.replace(k,v)
            if not overwrite and os.path.exists(target_fnfn): continue
            target_dir = os.path.dirname(target_fnfn)
            try:
                if not os.path.exists(target_dir): os.makedirs(target_dir)
            except FileExistsError:
                pass
            except FileNotFoundError:
                continue

            with zip_file.open(member) as source, open(target_fnfn,'wb') as target:
                shutil.copyfileobj(source, target)




def get_num_lines(filename):
    from smart_open import open

    def blocks(files, size=65536):
        while True:
            b = files.read(size)
            if not b: break
            yield b

    with open(filename, 'r', errors='ignore') as f:
        numlines=sum(bl.count("\n") for bl in blocks(f))

    return numlines




from ..imports import *









# loading txt/strings
def to_txt(txt_or_fn):
    # load txt
    if not txt_or_fn: return
    if os.path.exists(txt_or_fn):
        with open(txt_or_fn,encoding='utf-8',errors='replace') as f:
            txt=f.read()
    else:
        txt=txt_or_fn
    
    # clean?
    txt=txt.replace('--','—')

    return txt

    
    
def to_txt_from_scandf(scandf):
    return '\n\n'.join(
        '\n'.join(
            ''.join(
                linepartdf.index.get_level_values('line_str')[0]
                for linepart_i,linepartdf in linedf.groupby('linepart_i')
            )
            for line_i,linedf in stanzadf.groupby('line_i')
        )
        for stanza_i,stanzadf in scandf.groupby('stanza_i')
    )
    
def to_stanzas(full_txt):
    return [st.strip() for st in full_txt.strip().split('\n\n') if st.strip()]
def to_lines_str(stanza_txt):
    return [l.strip() for l in stanza_txt.split('\n') if l.strip()]
def to_sents_str(stanza_txt):
    return [
        sent.replace('\n',' ').strip()
        for sent in nltk.sent_tokenize(stanza_txt)
    ]

# def to_lineparts(linetxt,seps=set(',:;–—()[].!?'),min_len=1,max_len=25):
#     o=[]
#     for sent in to_sents_str(linetxt):
#         toks=tokenize_agnostic(sent)
#         ophrase=[]
#         for tok in toks:
#             ophrase+=[tok]
#             # print(tok,ophrases)
#             ophrase_len=len([x for x in ophrase if x[0].isalpha()])
#             if ophrase_len>=min_len and (tok in seps or ophrase_len>=max_len):
#                 o+=[''.join(ophrase)]
#                 ophrase=[]
#         # if ophrase:
#         #     if o and not any(x.isalpha() for x in ophrase):
#         #         o[-1]+=''.join(ophrase)
#         #     else:
#         #         o+=[''.join(ophrase)]
#         # if ophrase and any(x.isalpha() for x in ophrase): o+=[''.join(ophrase)]
#     return o    


def to_lineparts(linetxt,seps=set(',:;–—()[].!?'),min_len=1,max_len=10):
    o=[]
    for sent in to_sents_str(linetxt):
        sentparts=[]
        sentpart=[]
        for token in tokenize_nice(sent):
            pref,tok,suf = split_punct(token)

            # end prev?
            if sentpart and set(pref)&set(seps) and len(sentpart)>=min_len:
                sentparts.append(sentpart)
                sentpart=[]

            # add no matter what
            sentpart.append(token)

            # end after? or if too long?
            if sentpart and ((set(suf)&set(seps) and len(sentpart)>=min_len) or len(sentpart)>=max_len):
                sentparts.append(sentpart)
                sentpart=[]

        # add if remaining
        if sentpart:
            sentparts.append(sentpart)
            sentpart=[]
        o+=sentparts
    return [''.join(x) for x in o]

def to_words(line_txt,lang_code=DEFAULT_LANG):
    lang = to_lang(lang_code)
    o=lang.tokenize(line_txt)
    #print(o)
    #stop
    return o
    
def to_syllables(word_txt,lang_code=DEFAULT_LANG):
    return lang.syllabify(word_txt)

def scan(txt_or_fn,**kwargs):
    return setindex(
        pd.concat(
            scan_iter(txt_or_fn,**kwargs)
        )
    )

def do_scan_iter(obj,cache=False,**kwargs):
    (
        stanza_i,stanza_txt,
        line_i,line_txt,
        linepart_i,linepart_txt,
        #key
    ) = obj
    odf=line2df(linepart_txt, **kwargs)
    odf=odf[odf.syll_i!=0]


    try:
        assign_proms(odf)
    except KeyError:
        pass
    return odf
        

def scan_iter(txt_or_fn,lang=DEFAULT_LANG,
        progress=True,
        incl_alt=INCL_ALT,
        num_proc=DEFAULT_NUM_PROC,
        linebreaks=True,
        phrasebreaks=True,
        verse=None,
        prose=None,
        desc='Iterating over line scansions',
        **kwargs):
    full_txt=to_txt(txt_or_fn)
    if not full_txt: return
    
    if verse==True or prose==False:
        linebreaks=True
        phrasebreaks=False
    elif prose==True or verse==False:
        linebreaks=False
        phrasebreaks=True

    df=pd.DataFrame()
    dfl=[]
    to_lines_now = to_lines_str if linebreaks else to_sents_str
    kwargs['lang']=lang
    kwargs['incl_alt']=incl_alt
    
        
    objs=[
        (
            stanza_i,stanza_txt,
            line_i,line_txt,
            linepart_i,linepart_txt,
            # getkey(linepart_txt)
        )
        for stanza_i,stanza_txt in enumerate(to_stanzas(full_txt))
        for line_i,line_txt in enumerate(to_lines_now(stanza_txt))
        for linepart_i,linepart_txt in enumerate(
            to_lineparts(line_txt) if phrasebreaks else [line_txt]
        )
    ]    


    iterr=pmap_iter(
        do_scan_iter,
        objs,
        kwargs=kwargs,
        num_proc=num_proc,
        desc=desc,
        progress=progress
    )

#     dfl=[]
    i=0#
    #with get_db('lines','c') as db:
    lpi=None
    ol=[]
    for dfi,df in enumerate(iterr):
        (stanza_i,stanza_txt,line_i,line_txt,linepart_i,linepart_txt) = objs[dfi]
        if not linepart_txt: continue
        lpinow=(stanza_i,line_i)
        if lpi is not None and lpi!=lpinow and ol:
            yield pd.concat(ol)
            ol=[]
        
        df['stanza_i']=stanza_i+1
        df['line_i']=line_i+1
        df['line_str']=line_txt
        df['linepart_i']=linepart_i+1
        df['linepart_str']=linepart_txt

        df['linepart_num_syll']=len(df[df['word_ipa_i']==1])
        df['linepart_num_monosyll']=sum([
            1 if len(g)==1 else 0
            for i,g in df[df.word_ipa_i==1].groupby('word_i')
        ])

        #if 'is_syll' in set(df.columns):
        #   uniqdf=df[df.is_syll==1].drop_duplicates(['word_i','syll_i'])
        #   df['linepart_num_syll']=len(uniqdf)


        #display(df)
        #stop

        ol+=[df]
        lpi=lpinow
    if ol: yield pd.concat(ol)


def assign_proms(df):
    # set proms
    df['prom_stress']=pd.to_numeric(df['syll_ipa'].apply(getstress),errors='coerce')
    df['prom_weight']=pd.to_numeric(df['syll_ipa'].apply(getweight),errors='coerce')
    df['syll_stress']=df.prom_stress.apply(lambda x: {1.0:'P', 0.5:'S', 0.0:'U'}.get(x,''))
    df['syll_weight']=df.prom_weight.apply(lambda x: {1.0:'H', 0.0:'L'}.get(x,''))

    df['prom_strength']=[
        x
        for i,df_word in df.groupby(['word_i','word_ipa_i'])
        for x in getstrength(df_word)
    ]
    df['is_stressed']=(df['prom_stress']>0).apply(np.int32)
    df['is_unstressed']=(df['prom_stress']==0).apply(np.int32)
    df['is_heavy']=(df['prom_weight']==1).apply(np.int32)
    df['is_light']=(df['prom_weight']==0).apply(np.int32)
    df['is_peak']=(df['prom_strength']==1).apply(np.int32)
    df['is_trough']=(df['prom_strength']==0).apply(np.int32)


def get_info_byline(txtdf):
    linedf=txtdf.xs([0,0,0],level=['word_i','word_ipa_i','syll_i'])
    linedf=linedf.reset_index(level=['word_str','word_ipa','syll_str','syll_ipa','line_ii'],drop=True)
    return linedf#.reset_index().set_index(['stanza_i','line_i'])







# sonnet="""
# How heavy do I journey on the way,
# When what I seek, my weary travel's end,
# Doth teach that ease and that repose to say
# 'Thus far the miles are measured from thy friend!'
# The beast that bears me, tired with my woe,
# Plods dully on, to bear that weight in me,
# As if by some instinct the wretch did know
# His rider loved not speed, being made from thee:
# The bloody spur cannot provoke him on
# That sometimes anger thrusts into his hide;
# Which heavily he answers with a groan,
# More sharp to me than spurring to his side;
# For that same groan doth put this in my mind;
# My grief lies onward and my joy behind.
# """


# # def iter_windows(df,window_len=3):
# #     for linedf in iter_lines(df):
# #         for combodf in iter_combos(linedf):
# #             ldf=combodf.reset_index()
# #             ldf = ldf[ldf.word_ipa!=""]
# #             display(ldf)
# #             indices=list(range(len(ldf)))
# #             for indnow in slices(indices,window_len):
# #                 dfslice=combodf.iloc[indnow[0] : indnow[-1]+1]
# #                 display(dfslice)
# #                 stop
                

# # def iter_windows(df,window_len=3):
# #     for linedf in iter_lines(df):
# #         for combodf in iter_combos(linedf):
# #             ldf=combodf.reset_index()
# #             linedf_nopunc = ldf[ldf.word_ipa!=""]
# #             linewords=[y for x,y in sorted(linedf_nopunc.groupby('word_i'))]        
# #             for lwi,lineword_slice in enumerate(slices(linewords,window_len)):
# #                 for wi,word_combo in enumerate(apply_combos(pd.concat(lineword_slice), 'word_i', 'word_ipa_i', combo_key='word_combo_i')):
# #                     word_combo['word_window_i']=lwi
# #                     yield word_combo

            
            
# # #             for lwi,lineword_slice in enumerate(slices(linewords,window_len)):
# # #                 for wi,word_combo in enumerate(apply_combos(pd.concat(lineword_slice), 'word_i', 'word_ipa_i', combo_key='word_combo_i')):
# # #                     word_combo['word_window_i']=lwi
# # #                     yield word_combo



# def iter_parsed_windows(df):
#     for window in iter_windows(df):
#         for parsed in iter_parsed(window):
#             yield parsed


# def get_parsed_windows(txtdf):
#     # get all combos
#     return pmap_groups(
#         iter_parsed_windows,
#         txtdf.groupby(['stanza_i','line_i']),
#         progress=True,
#         num_proc=7,
#         desc='Parsing all windows'
#     )






from ..imports import *
from .txtparsing import *


"""
ITER LINES
"""

def iter_lines(txt_or_txtdf,num_proc=1,**kwargs):
    if type(txt_or_txtdf) == str:
        yield from scan_iter(txt_or_txtdf,num_proc=num_proc,**kwargs)
    else:
        for i,g in txt_or_txtdf.groupby(['stanza_i','line_i']):#,'linepart_i']):
            yield g

def iter_combos(txtdf,num_proc=1):
    for dfline in iter_lines(txtdf,num_proc=num_proc):
        # ldf=resetindex(dfline)
        ldf=dfline.reset_index()
        # linedf_nopunc = ldf[ldf.is_syll!=0]
        linedf_nopunc=ldf
        for dfi,dfcombo in enumerate(apply_combos(linedf_nopunc, 'word_i', 'word_ipa_i')):
            dfcombo['combo_i']=dfi
            dfcombo['combo_stress']=''.join(dfcombo.syll_stress)
            # dfcombo['combo_syll_i']=pd.Series(list(index_by_truth(dfcombo.is_syll)))+1
            dfcombo['combo_syll_i']=list(range(len(dfcombo)))
            dfcombo['combo_syll_i']+=1
            dfcombo['combo_num_syll']=dfcombo['combo_syll_i'].max()
            dfcombo['combo_ipa']=' '.join(
                '.'.join(wdf.syll_ipa)
                for wi,wdf in sorted(dfcombo.groupby('word_i'))
            )
            yield dfcombo
            #yield setindex(dfcombo)

def get_all_combos(txtdf):
    return resetindex(pd.concat(iter_combos(txtdf)))


"""
ITER PARSES
"""

def parse(txt_or_txtdf,*x,verbose=True,**y):
    """
    Return parses
    """
    o=list(parse_iter(txt_or_txtdf,*x, verbose=verbose, **y))
    odf=pd.concat(o) if len(o) else pd.DataFrame()
    odf=odf.sort_index()
    #if verbose:
    #    printm(f'{info_parses(odf)}')
    return odf


def parse_iter(
        txt_or_txtdf,
        by_line=True,
        by_syll=False,
        only_best=False,
        only_unbounded=True,
        num_proc=DEFAULT_NUM_PROC,
        progress=True,
        force=False,
        verbose=True,
        **kwargs):
    """
    Return parses as iter
    """
    if by_syll: by_line=False
    elif by_line: by_syll=False
#     if by_line: by_syll=False
#     if not by_syll: by_line=True
    yield from iter_parsed_lines(
        txt_or_txtdf,
        num_proc=num_proc,
        progress=progress,
        by_line=by_line,
        only_best=only_best,
        only_unbounded=only_unbounded,
        force=force,
        verbose=verbose,
        **kwargs
    )

def iter_parsed_lines(
        txt_or_txtdf,
        num_proc=DEFAULT_NUM_PROC,
        progress=False,
        by_line=False,
        only_best=False,
        only_unbounded=True,
        verbose=True,
        rebound_parses=False,
        verse=None,
        prose=None,
        **kwargs
        ):
    # preproc
    preproc_gen = iter_lines(
        txt_or_txtdf,
        progress=progress if not verbose else False,
        desc='Metrically parsing lines',
        verse=verse,
        prose=prose,
        **kwargs#.get('kwargs',{})
    )
    
    # proc
    kwargs['only_unbounded']=only_unbounded
    kwargs['engine']=ENGINE
    proc_gen = parmap(
        do_iter_parsed_lines,
        preproc_gen,
        N=num_proc,
        kwargs=kwargs
    )
    
    # postproc
    postproc_func = partial(
        postproc_dfline,
        by_line=by_line,
        only_best=only_best,
        # only_unbounded=only_unbounded,
        verbose=verbose,
        rebound_parses=rebound_parses,
        verse=verse,
        prose=prose,
        **kwargs
    )
    postproc_gen = map(postproc_func, proc_gen)
    
    # yield from total queue
    yield from postproc_gen

def do_iter_parsed_lines(
        linedf,
        num_proc=DEFAULT_NUM_PROC,
        force=True,
        cache=False,
        verbose=True,
        **kwargs):
    o=[]
    if not 'linepart_i' in set(linedf.columns) and not 'linepart_i' in set(linedf.index.names): return pd.DataFrame()
    for linepart_i,linepartdf in sorted(linedf.groupby('linepart_i')):
        if kwargs.get('engine')==ENGINE_PROSODIC:
            odf=parse_prosodic_line(
                linepartdf,
                **kwargs
            )
        else:
            o=list(iter_parsed_combos(
                linepartdf,
                num_proc=1,
                progress=False,
                **kwargs
            ))  
            odf=pd.concat(o) if len(o) else pd.DataFrame()
        o+=[odf]
    odf=pd.concat(o) if len(o) else pd.DataFrame()    
    return odf
    


def iter_parsed_combos(
        txtdf,
        num_proc=DEFAULT_NUM_PROC,
        progress=False,
        **kwargs):

    iter1=tqdm(
        iter_lines(txtdf,**kwargs),
        total=get_num_lines(txtdf),
        disable=(not progress)
    )

    for linedf in iter1:
        try:
            gdf=get_all_combos(linedf)
            yield from pmap_iter_groups(
                parse_combo,
                gdf.groupby(['stanza_i','line_i','combo_i']),
                num_proc=1,
                #progress=progress,
                kwargs=kwargs,
                progress=False,
                desc='Metrically parsing lines (+word IPA combinations)'
            )
        except ValueError:
            yield linedf







def parse_combo(linecombodf,engine=ENGINE,**kwargs):
    if engine==ENGINE_PROSODIC:
        odf=parse_prosodic_combo(linecombodf,**kwargs)
    else:
        raise Exception('Only prosodic implemented right now')
        # odf=parse_combo_cadence(linecombodf,**kwargs)
    return odf



def parse_prosodic_line(
        txt_or_txtdf,
        lang='en',
        meter='default_english',
        use_espeak=True,
        constraints=DEFAULT_CONSTRAINTS,
        constraint_weights=None,
        set_index=True,
        **kwargs):
    p.config['resolve_optionality']=1
    p.config['print_to_screen']=False
    p.config['en_TTS_ENGINE']='espeak' if use_espeak else 'none'
    p.config['num_bounded_parses_to_store']=NUM_BOUNDED_TO_STORE

    # print(constraints)
    # print(kwargs)
    
    meter_obj=p.config['meters'][meter]
    config=default_config=meter_obj.config


    if constraints is not None:
        constraints_pros=[
            f'{constraint_names_in_prosodic[c]}/{constraint_weights[i] if constraint_weights is not None else 1}'
            for c in constraints
        ]
        config['constraints']=constraints_pros
        meter_obj=p.Meter(config)
    
    txtdf = scan(txt_or_txtdf) if type(txt_or_txtdf)==str else txt_or_txtdf
    txtdf = txtdf.sort_index()
    
    line_words=[]
    plang=p.dict[lang]
    plang.getprep_syllabify_orth=None
    for wi0,wdf0 in txtdf.reset_index().groupby('word_i'):
        word_objs=[]
        for wi,wdf in wdf0.groupby('word_ipa_i'):
            ipa_str='.'.join(wdf.syll_ipa)
            sylls_text=list(wdf.syll_str)
            sylls_text_zp = [zero_punc(sx) for sx in sylls_text]
            token = wdf.iloc[0].word_str
            #print(wi0,wi,token,ipa_str)
            word_obj = plang.make((ipa_str,sylls_text_zp), token)
            word_obj.word_i=wi0
            word_obj.word_ipa_i=wi
            try:
                if wdf.is_funcword.iloc[0]==1: word_obj.feats['functionword']=True
            except AttributeError:
                display(wdf)
                stop
            # print(word_obj, word_obj.feats)
            word_objs.append(word_obj)
        wordtoken_obj = p.WordToken(word_objs, token)
        line_words.append(wordtoken_obj)
    l=p.Line()
    l.children = line_words
    l.parse(meter=meter_obj)
    odf=prosodic_line_to_data(l,txtdf,**kwargs)
    if 'parse_pos_i' in set(odf.columns): odf['parse_pos_i']+=1
    if 'parse_i' in set(odf.columns): odf['parse_i']+=1
    if 'combo_i' in set(odf.columns): odf['combo_i']+=1
    # if set_index: odf=setindex(odf,sort=True)
    return odf
        

def prosodic_line_to_data(
        line,
        txtdf,
        viols_in_parse_str=True,
        total_col=TOTALCOL,
        only_unbounded=True,
        posthoc_constraints=APPLY_POSTHOC_CONSTRAINTS,
        **kwargs):
    
    if line is None: return pd.DataFrame()
    if line.allParses() is None: return pd.DataFrame()
    all_parses = [(parse,True) for parse in line.allParses()]
    if not only_unbounded:
        all_parses+=[(parse,False) for parse in line.boundParses()][:NUM_BOUNDED_TO_STORE]
    needed_len=None
    O=[]
    for parse_i,(parse,is_unbounded) in enumerate(all_parses):
        o=[]
        parse_mstr=parse.str_meter()
        if needed_len is None:
            needed_len=len(parse_mstr)
        else:
            if len(parse_mstr)!=needed_len:
                break
        parse_str=parse.posString(viols=viols_in_parse_str)
        # parse_str=parse_str.replace('|',' ')
        # parse_mstr=' '.join(parse_mstr)
        bounded_by='' if is_unbounded else parse.boundedBy.str_meter()
        for pos_i,pos in enumerate(parse.positions):
            constraintd=dict(sorted(
                (constraint_names_from_prosodic.get(k.name,k.name),v)
                for k,v in pos.constraintScores.items()
            ))
            
            if total_col not in constraintd: constraintd[total_col]=sum(constraintd.values())
            mval=pos.meterVal
            mval2=''.join(mval for n in range(len(pos.slots)))
            
            nsyll=0
            for slot_i,slot in enumerate(pos.slots):
                nsyll+=1
                odx=dict(
                    word_i=slot.word.word_i,
                    word_ipa_i=slot.word.word_ipa_i,
                    syll_i=slot.wordpos[0],
                    parse_i=parse_i,
                    parse=parse_mstr,
                    parse_str=parse_str,
                    parse_pos_i=pos_i,
                    parse_pos=mval2,
                    parse_syll_i=nsyll,
                    parse_syll=mval,
                    **constraintd,
                    is_w=int(mval=='w'),
                    is_s=int(mval=='s'),
                    parse_num_pos=len(parse.positions),
                    parse_num_syll=len(parse_mstr.replace(' ','')),
                    parse_is_bounded=not is_unbounded,
                    parse_bounded_by=bounded_by
                )
                o.append(odx)
        if not len(o): continue
        odf=pd.DataFrame(o)
        
        
        odf=txtdf.reset_index().merge(
            odf,
            how='inner',
            on=['word_i','word_ipa_i','syll_i'],
            suffixes=['','_parse']
        )

        if posthoc_constraints:
            apply_posthoc_constraints(odf)
        
        


        O+=[odf]

    ## add back combo data
    if not len(O): return pd.DataFrame()
    odf=pd.concat(O)
    combos_sofar={}
    def tocombo(xdf):
        #display(xdf)
        key=tuple(xdf.reset_index().syll_ipa)
        if not key in combos_sofar: combos_sofar[key]=len(combos_sofar)
        return combos_sofar[key]
    odf=pd.concat(
        g.assign(
            combo_i=tocombo(g),
            combo_stress=' '.join(''.join(wdf.syll_stress) for wi,wdf in sorted(g.groupby('word_i'))),
            combo_ipa=' '.join('.'.join(wdf.syll_ipa) for wi,wdf in sorted(g.groupby('word_i')))
        )
        for i,g in odf.groupby('parse_i')
    )
    #print('3')
    #display(odf)

    return odf




def rank_parses(dfparses):
    odf=dfparses
    odf[TOTALCOL]=odf[[col for col in odf.columns if col.startswith('*') and col!=TOTALCOL]].sum(axis=1)
    odf['is_troch']=odf.parse.apply(lambda x: int(x and x[0]=='s'))
    newrankdf=odf[['parse_i',TOTALCOL]].groupby('parse_i').sum().sort_values(TOTALCOL)
    newrankdf['ranks']=newrankdf[TOTALCOL].rank(method='dense')
    newrank=dict(zip(newrankdf.index, newrankdf.ranks))
    odf['parse_rank']=odf.parse_i.apply(lambda x: int(newrank.get(x,0)))
    return odf

def postproc_dfline(
        dfline,
        by_line=False,
        verbose=True,
        verbose_bylinepart=False,
        only_best=False,
        only_unbounded=False,
        rebound_parses=False,
        verse=None,
        prose=None,
        **kwargs):
    #display(dfline)
    o=[]

    if not 'linepart_i' in set(dfline.columns) and not 'linepart_i' in set(dfline.index.names): return pd.DataFrame()
    for i,dfg in sorted(dfline.groupby('linepart_i')):
        dfg=rank_parses(dfg)
        if rebound_parses: dfg=bound_parses(dfg,only_unbounded=only_unbounded)
        if verbose and verbose_bylinepart and not verse: display(show_parse(dfg))
        o+=[dfg]
    if verbose and verbose_bylinepart and not verse: print()
    if verbose and (not verbose_bylinepart or verse):
        odf=pd.concat(o) if len(o) else pd.DataFrame()
        display(show_parse(odf))
    if by_line:
        o=[to_lines(dfg) for dfg in o]
    odf=pd.concat(o) if len(o) else pd.DataFrame()
    
    if only_best and ('parse_rank' in set(odf.index.names) or 'parse_rank' in set(odf.columns)):
        # odf=odf[odf.parse_rank==1]
        odf=odf.query('parse_rank==1')
    odf=setindex(odf)
    return odf

def is_bounded(pdf1,pdf2,totalcol=TOTALCOL,cprefix='*'):
    cdf1=pdf1[[c for c in pdf1.columns if c.startswith(cprefix) and c!=totalcol]]
    cdf2=pdf2[[c for c in pdf2.columns if c.startswith(cprefix) and c!=totalcol]]
    o=[]
    
    c1orig=cdf1.sum()
    c2orig=cdf2.sum()

    for c1,c2 in [(c1orig,c2orig), (c2orig,c1orig)]:
        greater=c2>=c1
        equal=c1==c2
        bounds=set(greater)=={True} and set(equal)!={True}
        #print(c1,c2,bounds,set(greater),set(equal))
        o+=[bounds]
    #stop
    return tuple(o)

def bound_parses(odforig,only_unbounded=False):
    odfg=odf=odforig#.reset_index()
    prank2meter=dict(zip(odf.parse_rank, odf.parse))
    if 'parse_is_bounded' in set(odf.columns):
        odf.parse_bounded_by=[
            ''.join(x) if type(x)==str and x else ''
            for x in odf.parse_bounded_by
        ]
        odfg=odf[odf.parse_is_bounded==False]
    else:
        odfg=odf.assign(parse_is_bounded=False)
    
    grps=sorted(list(odfg.groupby('parse_rank')),reverse=True)
    not_ok_ranks=set()
    bounded_by_d={}
    for parse_rank1,parsedf1 in grps:
        for parse_rank2,parsedf2 in grps:
            if parse_rank1>=parse_rank2: continue
            if parse_rank1 in not_ok_ranks or parse_rank2 in not_ok_ranks: continue
            x_bounds_y, y_bounds_x = is_bounded(parsedf1,parsedf2)
            if y_bounds_x: 
                not_ok_ranks|={parse_rank1}
                if parse_rank1 not in bounded_by_d:
                    bounded_by_d[parse_rank1]=prank2meter.get(parse_rank2)
            if x_bounds_y:
                not_ok_ranks|={parse_rank2}
                if parse_rank2 not in bounded_by_d:
                    bounded_by_d[parse_rank2]=prank2meter.get(parse_rank1)
    
    #odf=odf[~odf.index.get_level_values('parse_rank').isin(not_ok_ranks)]
    def isbound(isb,prank):
        if isb: return True
        if prank in not_ok_ranks: return True
        return False
    def isboundedby(bby,prank):
        if bby: return bby
        return bounded_by_d.get(prank,'')
        
    odf['parse_is_bounded']=[isbound(isb,prank) for isb,prank in zip(odf.parse_is_bounded, odf.parse_rank)]
    odf['parse_bounded_by']=[isboundedby(isb,prank) for isb,prank in zip(odf.parse_bounded_by.fillna(''), odf.parse_rank)]
    if only_unbounded: odf=odf[odf.parse_is_bounded==False]
    # odf.parse_rank=odf.parse_rank.rank(ascending=True,method='dense').apply(int)
    
    return odf






"""
Summarizing by line
"""

def to_lines(parses,totalcol=TOTALCOL,rankcol=PARSERANKCOL, agg=sum):
    # parses
    rparses=resetindex(parses)
    rcols=set(rparses.columns)
    pkcols = [c for c in PARSELINEKEY if c in rcols]
    # sum by actual data lines
    valcols = [c for c in rparses.select_dtypes('number').columns
              if not c in pkcols
              if not c.endswith('_i')
              if not c.endswith('_ii')
              if not c in {'index','level_0'}
              ]
    aggfunc = dict((vc,sum) for vc in valcols)
    # medians=['combo_num_syll','parse_num_syll','line_num_syll','parse_num_pos']
    medians = [col for col in rcols if '_num_' in col]
    for mx in medians:
        if mx in rcols:
            aggfunc[mx]=np.median
    lgby=['stanza_i','line_i','linepart_i']
    try:
        linesums = rparses.groupby(pkcols).agg(aggfunc)
    except Exception as e:
        print('!!',e)
        return pd.DataFrame()
    return linesums
    
def iter_parses(*x,by_line=True,**y):
    yield from iter_parsed_lines(*x,by_line=by_line,**y)
    
    
    
  

###### SHOWING


accentd={
    'a':'á',
    'e':'é',
    'i':'í',
    'o':'ó',
    'u':'ú',
    'y':'ý',
}
for k,v in list(accentd.items()): accentd[k.upper()]=v.upper()

def show_parses(parses_bysyll):
    try:
        from IPython.display import Markdown
        return Markdown(show_parses_md(parses_bysyll))
    except ImportError:
        return show_parses_txt(parses_bysyll)
    
def show_parse_md(dfparse):
#     print(dfparse.reset_index().line_i.nunique())
#     print(dfparse.reset_index().linepart_i.nunique())
    dfp=dfparse.reset_index()
    mdline=[]
    for l,dflp in sorted(dfp.groupby('linepart_i')):
        mdrow=[]
        #display(dflp)
        
        for w,dfword in sorted(dflp.groupby('word_i')):
            mdword=[]
            for s,dfsyll in sorted(dfword.groupby('syll_i')):
                row=dfsyll.iloc[0]
                wpref,w,wsuf = split_punct(row.syll_str)
                
                if row.is_stressed:
                    w2=[]
                    accented=False
                    numvowels=len([y for y in w if y in accentd])
                    for x in w:
                        if not accented and x in accentd:
                            if numvowels<2 or x!='y':
                                x=accentd[x]
                                accented=True
                        w2+=[x]
                    w=''.join(w2)
                    
                md=w#f'{row.syll_str}'
                if row.is_s: md=f'<b>{md}</b>'
                #if row.is_s: md=f'<u>{md}</u>'
                #if row.is_w: md=f'<i>{md}</i>'
                #if row.is_stressed: md=f'<i>{md}</i>'
                if row['*total']>0:
                    md=f'<span style="color:darkred">{md}</span>'
                    # md=f'<i>{md}</i>'
                md=f'{wpref}{md}{wsuf}'
                mdword+=[md]
            mdrow.append(''.join(mdword))
        mdline.append(''.join(mdrow))
    
    # spltr='\n * '
    # o=spltr + spltr.join(mdline)#.replace('</u> <u>',' ')
    o=''.join(mdline)#.replace('</u> <u>',' ')
    return o

def get_best_parse(dfparse):
    for i,g in dfparse.groupby(['parse_rank','is_troch','parse_i']): break
    return g

def get_best_parses(dfparses):
    o=[get_best_parse(g) for i,g in dfparses.groupby(['stanza_i','line_i','linepart_i'])]
    return pd.concat(o) if len(o) else pd.DataFrame()

def show_parse(dfparse,only_best=True):
    dfparse=get_best_parses(dfparse)

    try:
        from IPython.display import Markdown
        return Markdown(show_parse_md(dfparse))
    except ImportError:
        return show_parse_txt(dfparse)
    

def show_parses_md(parses_bysyll):
    
    grps=sorted(parses_bysyll.reset_index().groupby(
        ['stanza_i','line_i']#,'linepart_i','parse_rank']    
    ))
    mdl=[show_parse_md(dfparse) for i,dfparse in grps]
    omd='\n<br/>'.join(mdl)
    return omd
    
 
def info_parses(dfparses,lcols=['stanza_i','line_i','linepart_i']):
    # byline
    dfparses=dfparses.query('parse_is_bounded==False & parse_num_syll>=4')
    if 'syll_i' in set(dfparses.columns)|set(dfparses.index.names):
        dfparses=to_lines(dfparses)

    

    infod={}
    infod['num_lines_total']=dfparses.groupby(['stanza_i','line_i','linepart_i']).size().shape[0]
    
    # avg num sylls
    num_sylls = dfparses.groupby(lcols).parse_num_syll.mean()
    num_monosylls = dfparses.groupby(lcols).linepart_num_monosyll.mean()
    infod['num_sylls_avg'] = num_sylls.mean()

    infod['num_monosylls_avg'] = num_monosylls.mean()
    num_monosylls_per_10syll = num_monosylls / num_sylls * 10
    infod['num_monosylls_per_10syll_avg'] = num_monosylls_per_10syll.mean()
    

    # avg num parses per line?
    num_parses = dfparses.groupby(lcols).size()# / dfparses.groupby(lcols).parse_num_syll.mean()
    infod['num_parses_avg'] = num_parses.mean()
    num_parses_per_10syll = num_parses / num_sylls * 10
    infod['num_parses_per_10syll_avg'] = num_parses_per_10syll.mean()
    
    
    # avg viols?
    vkeys=[col for col in dfparses.columns if col.startswith('*')]
    for vk in vkeys:
        vk2='' if vk=='*total' else f'__{vk}'
        num_viols = dfparses.groupby(lcols)[vk].mean()
        infod[f'num_viols_avg{vk2}'] = num_viols.mean()
        num_viols_per_10syll = num_viols /num_sylls * 10
        infod[f'num_viols_per_10syll_avg{vk2}'] = num_viols_per_10syll.mean()

    return infod
    
def show_info_parses(infod):
    return f'''
* Metrical friction: **{infod["num_viols_per_10syll_avg"]:.02f}** viols per 10 sylls (**{infod["num_viols_avg"]:.02f}** per parse)
* Metrical ambiguity: **{infod["num_parses_per_10syll_avg"]:.02f}** parses per 10 sylls (**{infod["num_parses_avg"]:.02f}** per parse)
'''




def is_ok_parse(parse,maxS=2,maxW=2):
    return ('s'*(maxS+1)) not in parse and ('w'*(maxW+1)) not in parse


def possible_parses(window_len,maxS=2,maxW=2):
    poss = list(product(*[('w','s') for n in range(window_len)]))
    poss = [''.join(x) for x in poss]
    poss = [x for x in poss if is_ok_parse(x)]
    poss = [x for x in poss if len(x)==window_len]
    return poss

def possible_parses_recursive(window_len,starter=[],maxS=2,maxW=2,allow_overshooting=False,as_you_go=False):
    wtypes = [tuple(['w']*n) for n in range(1,maxW+1)]
    stypes = [tuple(['s']*n) for n in range(1,maxS+1)]
    if as_you_go and starter: yield starter
    if not starter:
        for typ in wtypes + stypes:
            yield from possible_parses_recursive(window_len,starter=[typ])
    else:
        lenn=sum(len(x) for x in starter)
        if lenn>window_len:
            if allow_overshooting:  # does not match syll count but overshoots it
                yield starter 
        elif lenn==window_len:  # exact match
            yield starter #tuple([tuple(x) for x in starter])
        else:
            if starter[-1][-1]=='s':
                for wtype in wtypes:
                    yield from possible_parses_recursive(window_len,starter=starter+[wtype])
            else:
                for stype in stypes:
                    yield from possible_parses_recursive(window_len,starter=starter+[stype])

def possible_metrical_positions(window_len,**y):
    return possible_parses_recursive(window_len,**y)


### Data by foot
def possible_metrical_feet(num_sylls,maxS=2,maxW=2):
    """
    Creates a hierarchical index of possible metrical parses
    """

    l = list(possible_parses_recursive(num_sylls,maxS=maxS,maxW=maxW))
    # return as data
    ld=[]
    for parse_i,parse in enumerate(l):
        tlen=sum(len(x) for x in parse)
        parse_str = ''.join([''.join(pos) for pos in parse])
        d={
            'parse':parse_str,
#             'pos':''.join(pos),
            'num_pos':len(parse),
            'num_syll':tlen
        }
        
        for pos_i,pos in enumerate(parse):
            d[f'mpos_{pos_i}']=''.join(pos)
        ld.append(d)
    df=pd.DataFrame(ld).fillna('')
    indices = [c for c in df.columns if c.startswith('mpos_')]
    return df.set_index(indices)



### Data by syllable
def possible_metrical_feet_bysyll(num_sylls,maxS=2,maxW=2):
    l = possible_parses_recursive(num_sylls,maxS=maxS,maxW=maxW)
    ld=[]
    for parse_i,parse in enumerate(l):
        num_syll=sum(len(x) for x in parse)
        num_pos=len(parse)
        parse_str = ''.join([''.join(pos) for pos in parse])
        parse_sofar=''
        parse_sofar_bysyll=''

        si=0
        for pos_i,pos in enumerate(parse):
            pos_str=''.join(pos)
            parse_sofar+=pos_str
            for syll_i,syll in enumerate(pos):
                parse_sofar_bysyll+=syll
                d={
                    'parse_i':parse_i,
                    'parse':parse_str,
                    # 'parse_sofar_bypos':parse_sofar,
                    # 'parse_sofar_bysyll':parse_sofar_bysyll,
                    'parse_num_pos':num_pos,
                    'parse_num_syll':num_syll,
                    'parse_pos_i':pos_i,
                    'parse_pos':pos_str,
                    'parse_syll_i':si,
                    'parse_syll':syll
                }
                si+=1
                ld.append(d)
    df=pd.DataFrame(ld)
    return df
### Data by foot
def possible_metrical_feet_bypos(num_sylls,maxS=2,maxW=2):
    """
    Creates a hierarchical index of possible metrical parses
    """

    l = possible_parses_recursive(num_sylls,maxS=maxS,maxW=maxW)
    ld=[]
    for parse_i,parse in enumerate(l):
        tlen=sum(len(x) for x in parse)
        parse_str = ''.join([''.join(pos) for pos in parse])
        for pos_i,pos in enumerate(parse):
            d={
                'parse_i':parse_i,
                'parse':parse_str,
                'parse_num_pos':len(parse),
                'parse_num_syll':tlen,
                'parse_pos_i':parse_pos_i,
                'parse_pos':''.join(pos)
            }
            ld.append(d)
    df=pd.DataFrame(ld)
    return df


POSSD={}
def get_poss_df(nsyll):
    global POSSD
    if not nsyll in POSSD: POSSD[nsyll]=possible_metrical_feet_bysyll(nsyll)
    return POSSD[nsyll]






"""
Metrical positions
"""

def iter_unique_metrical_positions(linecombodf):
    lcdf=linecombodf.reset_index()
    numsyll=len(lcdf)
    poss=[]
    poss_parses=possible_metrical_positions(numsyll)
    for parse in poss_parses:
        posi=0
        for i,pos in enumerate(parse):
            o=( (posi,posi+len(pos)), pos, i)
            poss+=[o]
            posi+=len(pos)
    
    for unique_mpos in sorted(list(set(poss))):
        indices,pos,posi = unique_mpos
        posdf = lcdf.iloc[indices[0] : indices[1]]
        posdf = pd.DataFrame(posdf)
        posdf['parse_syll'] = pos
        posdf['parse_pos'] = ''.join(pos)
        posdf['parse_pos_i'] = posi
        yield posdf


def iter_parsed_metrical_positions(linecombodf):
    for posdf in iter_unique_metrical_positions(linedf):
        yield parse_group(posdf)


def parse_group(df,constraints=DEFAULT_CONSTRAINTS):
    """
    Actual parsing, interface to constraints
    """

    dfcols=set(df.columns)
    if not 'is_w' in dfcols: df['is_w']=[np.int(x=='w') for x in df.parse_syll]
    if not 'is_s' in dfcols: df['is_s']=[np.int(x=='s') for x in df.parse_syll]
    dfpc=apply_constraints(df)
    lkeys=[c for c in df.columns if c not in set(dfpc.columns)]
    return df[lkeys].join(dfpc)


def get_parsed_metrical_positions(linecombodf):
    return setindex(pd.concat(iter_parsed_metrical_positions(linecombodf)))


"""
Positions -> Parses
"""

def parse_combo_cadence(linecombodf, **kwargs):
    lfpc=linecombodf.reset_index()
    # setup
    numsyll=len(lfpc)
    done=set()
    # parse just the positions
    dfpm = get_parsed_metrical_positions(lfpc)
    dfpmr=resetindex(dfpm)
    dfposs=get_poss_df(numsyll)

    # join with parsed positions
    dfjoin=setindex(dfpmr.merge(dfposs, left_on=['combo_syll_i','parse_pos_i','parse_syll','parse_pos'], right_on=['parse_syll_i','parse_pos_i','parse_syll','parse_pos']))
    
    # concat on way out
    o=[]
    for parse,parsedf in dfjoin.groupby(level=['combo_i','parse_i']):
        if not is_ok_parse(parse): continue
        pdf=resetindex(parsedf).drop_duplicates(['parse_syll_i','parse_syll','parse_pos_i','parse_pos'])
        if not is_valid_mpos_combo(pdf,numsyll): continue
        pdf['parse_str']='|'.join(
            '.'.join(wdf.apply(lambda row: row.syll_str.upper() if row.is_s else row.syll_str.lower(), axis=1))
            for wi,wdf in sorted(pdf.groupby('parse_pos_i'))
        )
        o+=[pdf]
    odf=setindex(pd.concat(o))
    return odf

def is_valid_mpos_combo(parsedf,numsyll):
    if len(set(parsedf.combo_syll_i))!=numsyll: return False
    return True

def get_num_lines(txt_or_txtdf,splitter='\n',groupby=['stanza_i','line_i']):
    if txt_or_txtdf is None:
        return np.nan
    if type(txt_or_txtdf)==str:
        return len([x for x in txt_or_txtdf.split(splitter) if x.strip()])
    elif type(txt_or_txtdf)==pd.DataFrame:
        return len(txt_or_txtdf.groupby(groupby))
    return np.nan

def get_num_stanzas(txt_or_txtdf):
    return get_num_lines(txt_or_txtdf,splitter='\n\n',groupby='stanza_i')



def parse_line(linedf,engine=ENGINE,**kwargs):
    """
    Line DF -> Combos -> Parsed -> Concat
    """
    if engine==ENGINE_PROSODIC:
        return parse_prosodic_line(linedf,**kwargs)
    else:
        return setindex(pd.concat(parse_combo(combo) for combo in iter_combos(linedf)))






def apply_constraints_cadence(mpos_window,constraints=DEFAULT_CONSTRAINTS):
    total=None
    assert len(set(mpos_window.parse_pos_i))==1
    dfc=pd.DataFrame(index=mpos_window.index)
    for cname,cfunc in constraints.items():
        cvals=cfunc(mpos_window)
        dfc['*'+cname]=cvals
    dfc['*total']=dfc.sum(axis=1)
    return dfc




# default constraints
DEFAULT_CONSTRAINTS_FUNC = {
    'w/peak':no_weak_peaks,
    'w/stressed':no_stressed_weaks,
    's/unstressed':no_unstressed_strongs,
    # 's/trough':no_strong_troughs,
    'clash':no_clash,
    'lapse':no_lapse,
    'f-res':f_resolution,
    'w-res':w_resolution,
}




# sys imports
import os,sys
sys.path.append('/home/ryan/github/prosodic')
import warnings
warnings.filterwarnings('ignore')
import prosodic as p

from tqdm import tqdm
import pandas as pd,numpy as np,random,json,pickle,shutil
from collections import defaultdict,Counter
import subprocess,multiprocessing as mp
mp.set_start_method('fork')
from pprint import pprint
from itertools import product
pd.options.display.max_columns=False
import re,nltk
from sqlitedict import SqliteDict
import logging
logging.Logger.manager.loggerDict['sqlitedict'].disabled=True

from parmapper import parmap
from functools import partial


# constants
ENGINE_PROSODIC='prosodic'
ENGINE_CADENCE='cadence'
ENGINE=ENGINE_PROSODIC

MIN_WORDS_IN_PHRASE=2
DEFAULT_LANG='en'
PATH_HERE=os.path.abspath(os.path.dirname(__file__))
PATH_CODE=PATH_HERE
PATH_REPO=os.path.abspath(os.path.join(PATH_CODE,'..'))
PATH_HOME=os.path.join(os.path.expanduser('~'),'.cadence')
PATH_DATA=os.path.join(PATH_HOME,'data')
# PATH_DATA=os.path.join(PATH_REPO,'data')
DATA_URL='https://www.dropbox.com/s/fywmqrlpemjf43c/data_cadence.zip?dl=1'


PATH_NOTEBOOKS=os.path.join(PATH_REPO,'notebooks')
PATH_IPA_FEATS=os.path.join(PATH_DATA,'data.feats.ipa.csv')
INCL_ALT=True
DEFAULT_NUM_PROC=1
#mp.cpu_count()//2# - 1

KEEP_BEST=1
SBY=csby=['combo_i','word_i','syll_i']
PARSERANKCOL='parse_rank'

LINEKEY=[
    'stanza_i',
    'line_i',
    'line_str',
    'linepart_i',#'linepart_str',
    'linepart_str',
    'linepart_end_str',
    
    
    PARSERANKCOL,
    #'parse_str',
    'is_troch',
    'parse_i',
    'parse',
    # 'combo_stress',
    'parse_str',

    
    'combo_stress',
    'combo_ipa',
    'combo_i',
#     'combo_i','combo_stress','combo_ipa',
    'parse_is_bounded','parse_bounded_by',
    'parse_pos_i','parse_pos',
    'word_i','word_str','word_ipa_i','word_ipa',
    'syll_i','combo_syll_i','syll_str','syll_ipa','syll_stress','syll_weight',
    'parse_syll_i','parse_syll',
]
PARSELINEKEY = LINEKEY[:LINEKEY.index('parse_pos_i')]
PARSESYLLKEY=LINEKEY
TOTALCOL='*total'

DEFAULT_CONSTRAINTS = {
    '*w/peak',
    '*w/stressed',
    # '*s/unstressed',
    '*f-res',
    '*w-res'
}

NUM_BOUNDED_TO_STORE = 10

constraint_names_in_prosodic = {
    '*f-res':'footmin-f-resolution',
    '*s/unstressed':'stress.s=>-u',
    '*w-res':'footmin-w-resolution',
    '*w/peak':'strength.w=>-p',
    '*w/stressed':'stress.w=>-p',
    '*s/trough':'strength.s=>-u',
}
constraint_names_from_prosodic = dict((v,k) for k,v in constraint_names_in_prosodic.items())


txt="""In Xanadu did Kubla Khan
A stately pleasure-dome decree:
Where Alph, the sacred river, ran
Through caverns measureless to man
   Down to a sunless sea.
"""
line=txt.split('\n')[0]

APPLY_POSTHOC_CONSTRAINTS=False


# local imports
from .tools import *
from .langs import *
from .constraints import *
from .parsers import *
from .cadence import *


# check
check_basic_config()


from ..imports import *
from .english import scan as en_scan
CODE2LANG = {
    'en':en_scan
}

def to_lang(lang_code=None):
    global LANGS
    return LANGS[lang_code]


D_IPA=None
def get_d_ipa(fn=PATH_IPA_FEATS):
    global D_IPA
    if D_IPA is None: D_IPA=pd.read_csv(fn).set_index('ipa').to_dict()
    return D_IPA



def to_phons(syll_ipa):
    pass

def line2df(line_txt,
        lang=DEFAULT_LANG,
        sby=['word_i','word_ipa_i','syll_i'],
        incl_alt=INCL_ALT,**y):
    
    func = CODE2LANG.get(lang, CODE2LANG[DEFAULT_LANG] )
    odf=func(line_txt,incl_alt=incl_alt,**y)
    # try:
    #     # odf=odf.sort_values(sby)[
    #     #     sby + [col for col in odf.columns if col not in set(sby)]
    #     # ]
    #     # odf['is_syll']=odf.syll_ipa.apply(lambda x: int(x==''))#[int(x) for x in odf['syll_ipa']!='']
    #     #odf['line_num_syll']=len(odf[(odf.is_syll==1) & (odf.word_ipa_i==1)])
    #     #uniqdf=odf[odf.is_syll==1].drop_duplicates(['word_i','syll_i'])
    #     #odf['line_num_syll']=len(uniqdf)
    # except KeyError as e:
    #     pass
    return odf




# ## langs


# stress
def getstress(sylipa):
    if not sylipa.strip(): return ' '
    if sylipa.startswith("'"): return 1.0#""
    elif sylipa.startswith("`"): return 0.5
    return 0.0

def getstress_str(sylipa):
    x=getstress(sylipa)
    if x==1.0: return 'P'
    if x==0.5: return 'S'
    if x==0.0: return 'U'
    return ''

def getweight(sylipa):
    dipa=get_d_ipa()
    phondata = [dipa[ipastr] for ipastr in sylipa if ipastr in dipa]
    if not len(phondata): return np.nan
    ends_with_cons=phondata[-1]['cons']==True
    has_long_vowel=any(x['long']==True for x in phondata)
    is_dipthong=None #@TODO
    weight=1 if ends_with_cons or has_long_vowel or is_dipthong else 0
    return weight

def getweight_str(sylipa):
    if x==1: return 'H'
    if x==0: return 'L'
    return ''

# def getweight(self,df_syll):
#     ends_with_cons=df_syll.sort_values('phon_i')['phon_cons'].iloc[-1]==True
#     has_long_vowel=any(x==True for x in df_syll['phon_long'])
#     # print(has_long_vowel,ends_with_cons,[df_syll.sort_values('phon_i')['phon_cons'].iloc[-1]])
#     is_dipthong=None #@TODO
#     weight='H' if ends_with_cons or has_long_vowel or is_dipthong else 'L'
#     return [weight for i in range(len(df_syll))]

def getstrength(df_word):
    stresses=dict(zip(df_word.syll_i,df_word.prom_stress))
    strengths={}
    for si,(syll_i,syl) in enumerate(sorted(stresses.items())):
        prv=stresses.get(syll_i-1)
        nxt=stresses.get(syll_i+1)
        if nxt!=None and prv!=None:
            if syl>nxt or syl>prv:
                strength=1.0
            elif syl<nxt or syl<prv:
                strength=0.0
            else:
                strength=np.nan
        elif prv==None and nxt!=None:
            if syl>nxt:
                strength=1.0
            elif syl<nxt:
                strength=0.0
            else:
                strength=np.nan
        elif prv!=None and nxt==None:
            if syl>prv:
                strength=1.0
            elif syl<prv:
                strength=0.0
            else:
                strength=np.nan
        elif prv==None and nxt==None:
            strength=np.nan
        else:
            raise Exception("How? -getstrength()")
        #strengths.append(strength)
        strengths[syll_i]=strength
    return [
        strengths.get(syll_i)
        for syll_i in df_word.syll_i
    ]




from ..imports import *

"""
All constraints take the entire dataframe
"""

def no_stressed_weaks(df_mpos):
    return np.multiply(df_mpos.is_w , df_mpos.is_stressed)

def no_unstressed_strongs(df_mpos):
    return np.multiply(df_mpos.is_s, df_mpos.is_unstressed)
    
def no_weak_peaks(df_mpos):
    return np.multiply(df_mpos.is_w, df_mpos.is_peak)
def no_strong_troughs(df_mpos):
    return np.multiply(df_mpos.is_s, df_mpos.is_trough)

def w_resolution(df_mpos,weight=1):
    if len(set(df_mpos.word_i))<2: return np.nan # only applies to word-boundaries    
    # does this apply to weak positions?
    if df_mpos.parse_syll.iloc[0]=='w': return np.nan # cannot apply to strong positions

    firstsyll_islight=df_mpos.iloc[0].is_light
    firstsyll_isstressed=df_mpos.iloc[0].is_stressed
    if not (firstsyll_islight and firstsyll_isstressed): return weight
    return 0

def f_resolution(df_mpos,weight=1):
    if len(set(df_mpos.word_i))<2: return np.nan # only applies to word-boundaries
    if df_mpos.parse_pos.iloc[0]=='ss': return weight # cannot apply to strong positions
    if sum(df_mpos.is_funcword)!=len(df_mpos): return weight
    return 0


def no_window(s,badwindow=(1,1)):
    wlen=len(badwindow)
    l=[]
    window=[]
    for x in s:
        window.append(x)
        l.append(int(tuple(window)==badwindow))
        if len(window)>=wlen:window.pop(0)
    return l

def no_clash(dfparse):
    return no_window(dfparse.is_stressed,badwindow=(1,1))
def no_lapse(dfparse):
    return no_window(dfparse.is_stressed,badwindow=(0,0,0))
def no_nonalt(dfparse):
    ww=no_window(dfparse.is_w,badwindow=(1,1))
    ss=no_window(dfparse.is_s,badwindow=(1,1))
    return [1 if x or y else 0 for x,y in zip(ww,ss)]
def no_trochsub(dfparse):
    x=int(dfparse.is_s.iloc[0]==1)
    return [x]+[0 for y in range(len(dfparse)-1)]

def apply_posthoc_constraints(dfparse):
    dfparse['*clash']=[
        x
        for pi,pdf in dfparse.groupby('parse_i')
        for x in no_clash(pdf)
    ]
    dfparse['*lapse']=[
        x
        for pi,pdf in dfparse.groupby('parse_i')
        for x in no_lapse(pdf)
    ]
    dfparse['*nonalt']=[
        x
        for pi,pdf in dfparse.groupby('parse_i')
        for x in no_nonalt(pdf)
    ]


    dfparse[TOTALCOL]=dfparse[[col for col in dfparse.columns if col.startswith('*') and col!=TOTALCOL]].sum(axis=1)

